/*
 * Copyright (c) 2012 Flavio Ribeiro <fr@lps.usp.br>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

Qt.include('mathfunc.js');
Qt.include('seedrandom.js');
Qt.include('storage.js');

var alt_l = false;
var alt_r = false;

var stack = [];
var thisX = '';
var radix = '.';
var exp10 = 'e';

var CalculatorState = { Reset : 0,
                        Typing : 1,
                        DispSetup : 2,
                        AlphaEntrySTO: 3,
                        AlphaEntryRCL: 4,
                        AlphaSwapX: 5,
                        AlphaSwapY: 6,
                        ComplexFunction: 7,
                        HyperbolicFunction: 8 }
var state = CalculatorState.Reset;

var history_push = [];
var history_pop  = [];
var history_last = -1;

var DisplayOptions = { Automatic : 0, Fixed : 1, Scientific : 2, Engineering : 3 };
var AngleOptions   = { Radians : 0, Degrees : 1 };
var DecimalOptions = { Period : 0, Comma : 1 };
var BaseOptions    = { Decimal : 0, Hexadecimal : 1, Octal : 2, Binary : 3 };

var key_handler;

var memory = { sums_n      : 0,
               sums_x      : 0,
               sums_y      : 0,
               sums_xy     : 0,
               sums_x2     : 0,
               sums_x2     : 0,
               qmax        : 4096 };

var option = { display_width       : 12,     // display width in characters
               display_prec        : 5,
               display_rational    : false,
               display_mode        : DisplayOptions.Automatic,
               display_mode_new    : DisplayOptions.Automatic,
               angle_unit          : AngleOptions.Radians,
               decimal_separator   : DecimalOptions.Period,
               integer_base        : BaseOptions.Decimal };

function initialize()
{
    memory_load_AZ();
    memory_load_sums();
    memory_load_options();
    initialize_callbacks();
    update_display();
}

function set_option(name, value, redraw)
{
    if (name === 'integer_base' && option['integer_base'] !== value)
        option['integer_base_last'] = option['integer_base'];

    option[name] = value;
    setSetting('option_' + name, value);
    if (redraw)
        update_display();
}

function get_option(name)
{
    var x = getSetting('option_' + name);
    option[name] = (x !== null) ? parseInt(x) : option[name];
}

function set_memory(name, value)
{
    memory[name] = value;
    setSetting('memory_' + name, value);
}

function get_memory(name)
{
    var x = getSetting('memory_' + name);
    memory[name] = (x !== null) ? parseFloat(x) : 0.0;
}

function memory_load_options()
{
    get_option('display_prec');
    get_option('display_mode');
    get_option('angle_unit');
    get_option('decimal_separator');
    get_option('integer_base');
}

function memory_load_AZ()
{
    for (var i = 'A'.charCodeAt(0); i <= 'Z'.charCodeAt(0); i++)
        get_memory(String.fromCharCode(i))
}

function memory_clear_AZ()
{
    for (var i = 'A'.charCodeAt(0); i <= 'Z'.charCodeAt(0); i++)
       set_memory(String.fromCharCode(i), 0);
}

function memory_load_sums()
{
    get_memory('sums_n');
    get_memory('sums_x');
    get_memory('sums_y');
    get_memory('sums_xy');
    get_memory('sums_x2');
    get_memory('sums_y2');
}

function memory_save_sums()
{
    set_memory('sums_n',  memory['sums_n']);
    set_memory('sums_x',  memory['sums_x']);
    set_memory('sums_y',  memory['sums_y']);
    set_memory('sums_xy', memory['sums_xy']);
    set_memory('sums_x2', memory['sums_x2']);
    set_memory('sums_y2', memory['sums_y2']);
}

function memory_clear_sums()
{
    set_memory('sums_n',  0);
    set_memory('sums_x',  0);
    set_memory('sums_y',  0);
    set_memory('sums_xy', 0);
    set_memory('sums_x2', 0);
    set_memory('sums_y2', 0);
}

function memory_clear_stack()
{
    stack = [];
    history_reset();
}

function print_error()
{
    error_area.visible = true;
    error_area_timer.running = true;
}

function is_real()
{
    return option['integer_base'] === BaseOptions.Decimal;
}

function in_radians()
{
    return option['angle_unit'] === AngleOptions.Radians;
}

function argument_check(args)
{
    if ((state === CalculatorState.Typing && !enter_quiet()) ||
        (state === CalculatorState.Reset && stack.length < args) ||
        (state !== CalculatorState.Reset))
    {
        print_error();
        return false;
    }
    else
        return true;
}

function update_display()
{
    var display_string = '';

    switch(option['display_mode'])
    {
        case DisplayOptions.Automatic:
            disp.text = "auto";
            break;
        case DisplayOptions.Fixed:
            disp.text = "fixed";
            break;
        case DisplayOptions.Scientific:
            disp.text = "sci";
            break;
        case DisplayOptions.Engineering:
            disp.text = "eng";
            break;
    }

    switch(option['angle_unit'])
    {
        case AngleOptions.Radians:
            mode.text = "rad";
            break;
        case AngleOptions.Degrees:
            mode.text = "deg";
            break;
    }

    switch(option['integer_base'])
    {
        case BaseOptions.Decimal:
            base.text = "dec";
            break;
        case BaseOptions.Hexadecimal:
            base.text = "hex";
            break;
        case BaseOptions.Octal:
            base.text = "oct";
            break;
        case BaseOptions.Binary:
            base.text = "bin"
            break;
    }

    for (var i = 0; i < stack.length; i++)
    {
        switch(option['integer_base'])
        {
            case BaseOptions.Decimal:
                if (option['display_rational'])
                {
                    var q = Math.bestq(stack[i], memory['qmax']);
                    var p = Math.round(stack[i]*q);
                    display_string += '<br>' + p + '/' + q;
                }
                else
                {
                    switch(option['display_mode'])
                    {
                        case DisplayOptions.Automatic:
                            display_string += '<br>' + stack[i].toStringAuto();
                            break;
                        case DisplayOptions.Fixed:
                            display_string += '<br>' + stack[i].toStringFixed(option['display_prec']);
                            break;
                        case DisplayOptions.Scientific:
                            display_string += '<br>' + stack[i].toStringSci(option['display_prec']);
                            break;
                        case DisplayOptions.Engineering:
                            display_string += '<br>' + stack[i].toStringEng(option['display_prec']);
                            break;
                    }
                }
                break;
            case BaseOptions.Hexadecimal:
                display_string += '<br>' + stack[i].toStringUnsigned(16);
                break;
            case BaseOptions.Octal:
                display_string += '<br>' + stack[i].toStringUnsigned(8);
                break;
            case BaseOptions.Binary:
                display_string += '<br>' + stack[i].toStringUnsigned(2);
                break;
        }
    }

    if (state === CalculatorState.Typing)
        display_string += '<br>' + thisX + '_';

    if (option['decimal_separator'] === DecimalOptions.Comma)
        display_string = display_string.replace(/\./g, ',');

    if (state === CalculatorState.DispSetup)
    {
        switch(option['display_mode_new'])
        {
            case DisplayOptions.Fixed:
            case DisplayOptions.Scientific:
            case DisplayOptions.Engineering:
                display_string += '<br># decimals?';
                break;
        }
    }
    else if (state === CalculatorState.AlphaEntrySTO)
        display_string += '<br>Memory slot?';
    else if (state === CalculatorState.AlphaEntryRCL)
        display_string += '<br>Memory slot?';
    else if (state === CalculatorState.AlphaSwapX || state === CalculatorState.AlphaSwapY)
        display_string += '<br>Memory slot?';
    else if (state === CalculatorState.ComplexFunction)
        display_string += '<br>CPLX function?';
    else if (state === CalculatorState.HyperbolicFunction)
        display_string += '<br>HYP function?';

    var display_font = Math.floor(600/option['display_width']);
    if (display.font.pixelSize !== display_font)
        display.font.pixelSize = display_font;
    display.text = display_string.replace('<br>','');       // remove initial <br>

    display_flickable.contentX = display.paintedWidth - display_flickable.width;
    display_flickable.contentY = display.paintedHeight - display_flickable.height;
}

function enter_quiet()
{
    if (state === CalculatorState.Reset && stack.length > 0)
    {
        history_update([], [stack.top()]);
        stack.push(stack.top());
        return true;
    }
    else if (state === CalculatorState.Typing)
    {
        var n;

        if (thisX.length === 0)
        {
            state = CalculatorState.Reset;
            return true;
        }

        switch(option['integer_base'])
        {
            case BaseOptions.Decimal:
                n = parseFloat(thisX);
                break;
            case BaseOptions.Hexadecimal:
                n = parseInt(thisX, 16);
                break;
            case BaseOptions.Octal:
                n = parseInt(thisX, 8);
                break;
            case BaseOptions.Binary:
                n = parseInt(thisX, 2);
                break;
        }

        if (!isNaN(n))
        {
            history_update([], [n]);
            stack.push(n);
            state = CalculatorState.Reset;
            thisX = '';
            return true;
        }
    }

    return false;
}

function enter()
{
    if (!enter_quiet())
        print_error();
}

function clear()
{
    thisX = "";
    state = CalculatorState.Reset;
}

function change_sign()
{
    if (state === CalculatorState.Reset && stack.length > 0)
    {
        history_update([stack.top()], [-stack.top()]);
        stack[stack.length-1] = -stack.top();
    }
    else if (state === CalculatorState.Typing && thisX.length > 0)
    {
        var exp10_split = thisX.split(exp10);
        var exp10_count = exp10_split.length - 1;
        var number = (exp10_count === 0) ? exp10_split[0] : exp10_split[1];

        if (number.charAt(0) === '-')
            number = number.substring(1);
        else if (number.charAt(0) === '+')
            number = '-' + number.substring(1);
        else
            number = '-' + number;

        thisX = (exp10_count === 0) ? number : (exp10_split[0] + exp10 + number);
    }
    else
        print_error();
}

function append_exp10()
{
    if (state === CalculatorState.Typing && thisX.length > 0)
    {
        var exp10_count = thisX.split(exp10).length - 1;
        if (exp10_count === 0)
            thisX += exp10;
        else
            print_error();
    }
    else
        print_error();
}

function valid_digit(d)
{
    if (d.length !== 1)
        return false;

    var code = d.charCodeAt(0);
    switch(option['integer_base'])
    {
        case BaseOptions.Decimal:
            return ('0'.charCodeAt(0) <= code && code <= '9'.charCodeAt(0));
        case BaseOptions.Hexadecimal:
            return ('0'.charCodeAt(0) <= code && code <= '9'.charCodeAt(0)) ||
                   ('A'.charCodeAt(0) <= code && code <= 'F'.charCodeAt(0));
        case BaseOptions.Octal:
            return ('0'.charCodeAt(0) <= code && code <= '7'.charCodeAt(0));
        case BaseOptions.Binary:
            return ('0'.charCodeAt(0) <= code && code <= '1'.charCodeAt(0));
    }

    return false;
}

function append_digit(d)
{
    if (state === CalculatorState.Reset || state === CalculatorState.Typing)
    {
        if (thisX.length >= option['display_width'] || !valid_digit(d))
            print_error();      // prevents a number larger than the display or an invalid digit
        else
            thisX += d;

        state = CalculatorState.Typing;
    }
    else if (state === CalculatorState.DispSetup)
    {
        set_option('display_prec', parseInt(d), false);
        set_option('display_mode', option['display_mode_new'], false);
        state = CalculatorState.Reset;
    }
    else
        print_error();
}

function append_period()
{
    if (state === CalculatorState.Reset || state === CalculatorState.Typing)
    {
        var radix_count = thisX.split(radix).length - 1;
        var exp10_count = thisX.split(exp10).length - 1;
        if (radix_count === 0 && exp10_count === 0)
            thisX += radix;
        else
            print_error();

        state = CalculatorState.Typing;
    }
    else
        print_error();
}

function backspace()
{
    if (state === CalculatorState.Reset && stack.length > 0)
        history_update([stack.pop()], []);
    else if (state === CalculatorState.Typing)
        thisX = thisX.substring(0, thisX.length-1);
    else
        print_error();
}

function open_menu(mode)
{
    if (!argument_check(0))
        return;

    switch(mode) {
        case 'base':
            base_menu.open();
            break;

        case 'word':
            prec_menu.open();
            break;

        case 'disp':
            disp_menu.open();
            break;

        case 'mode':
            mode_menu.open();
            break;
    }
}

function rotate_down()
{
    if (!argument_check(0) || stack.length === 0)
        return;

    var t = stack.top();
    for (var i = stack.length-1; i >= 0; i--)
        stack[i] = stack[i-1];
    stack[0] = t;

    history_reset();        // reset history because we don't know how to undo this
}

function rotate_up()
{
    if (!argument_check(0) || stack.length === 0)
        return;

    var b = stack[0];
    for (var i = 0; i < stack.length-1; i++)
        stack[i] = stack[i+1];
    stack[stack.length-1] = b;

    history_reset();        // reset history because we don't know how to undo this
}

function history_reset()
{
    history_push = [];
    history_pop  = [];
    history_last = -1;
}

function history_update(last_pop, last_push)
{
    history_last++;
    history_pop  = history_pop.slice (0, history_last);
    history_push = history_push.slice(0, history_last);

    history_pop.push(last_pop);
    history_push.push(last_push);
}

function history_undo()
{
    if (!argument_check(0))
        return;

    if (history_last >= 0)
    {
        stack.splice(stack.length - history_push[history_last].length, history_push[history_last].length);
        stack = stack.concat(history_pop[history_last]);
        history_last--;
    }
    else
        print_error();
}

function history_redo()
{
    if (!argument_check(0))
        return;

    if ((history_last + 1) < history_push.length &&
        (history_last + 1) < history_pop.length)
    {
        history_last++;
        stack.splice(stack.length - history_pop[history_last].length, history_pop[history_last].length);
        stack = stack.concat(history_push[history_last]);
    }
    else
        print_error();
}

function lastx()
{
    if (!argument_check(0))
        return;

    if (history_last < 0 || history_pop[history_last].length === 0)
        print_error();
    else
        stack.push(history_pop[history_last].top());
}

function lastb()
{
    if (!argument_check(0))
        return;

    if (option['integer_base_last'] === undefined)
        print_error();
    else
        set_option('integer_base', option['integer_base_last'], false);
}

// Pearson correlation coefficient
function stats_r()
{
    var num = memory['sums_n']*memory['sums_xy'] - memory['sums_x']*memory['sums_y'];
    var den = Math.sqrt(memory['sums_n']*memory['sums_x2'] - Math.pow(memory['sums_x'],2)) *
              Math.sqrt(memory['sums_n']*memory['sums_y2'] - Math.pow(memory['sums_y'],2));
    return num/den;
}

// least-squares y-intercept
function stats_b()
{
    var num = memory['sums_y']*memory['sums_x2'] - memory['sums_x']*memory['sums_xy'];
    var den = memory['sums_n']*memory['sums_x2'] - Math.pow(memory['sums_x'],2);
    return num/den;
}

// least-squares slope
function stats_m()
{
    var num = memory['sums_n']*memory['sums_xy'] - memory['sums_x']*memory['sums_y'];
    var den = memory['sums_n']*memory['sums_x2'] - Math.pow(memory['sums_x'],2);
    return num/den;
}

function op(ret_count, arg_count, operator)
{
    var i;

    if (!argument_check(arg_count))
        return false;

    var args = (arg_count === 0) ? [] : stack.slice(-arg_count);
    var rets = [];
    var success = false;

    var args_copy = is_real() ? args.slice(0) : args.slice(0).map(Math.toInt).map(Math.truncateInt);

    if (ret_count === 0 && arg_count === 1)
        success = op01(rets, args_copy, operator);
    else if (ret_count === 1 && arg_count === 0)
        success = op10(rets, args_copy, operator);
    else if (ret_count === 1 && arg_count === 1)
        success = op11(rets, args_copy, operator);
    else if (ret_count === 0 && arg_count === 2)
        success = op02(rets, args_copy, operator);
    else if (ret_count === 1 && arg_count === 2)
        success = op12(rets, args_copy, operator);
    else if (ret_count === 2 && arg_count === 2)
        success = op22(rets, args_copy, operator);
    else if (ret_count === 2 && arg_count === 4)
        success = op24(rets, args_copy, operator);

    if (!is_real())
        rets = rets.map(Math.toInt).map(Math.truncateInt);

    if (!success)
    {
        print_error();
        return false;
    }

    history_update(args, rets);
    stack.splice(stack.length - arg_count, arg_count);
    stack = stack.concat(rets);
    return true;
}

// 1 return value
// 0 arguments
function op10(rets, args, operator)
{
    var x, num, den;

    switch(operator)
    {
        case 'pi':
            x = Math.PI;
            break;
        case 'rand':
            x = Math.random();
            break;
        case 'stats_n':
            x = memory['sums_n'];
            break;
        case 'stats_x':
            x = memory['sums_x'];
            break;
        case 'stats_y':
            x = memory['sums_y'];
            break;
        case 'stats_xy':
            x = memory['sums_xy'];
            break;
        case 'stats_x2':
            x = memory['sums_x2'];
            break;
        case 'stats_y2':
            x = memory['sums_y2'];
            break;
        case 'stats_mx':
            x = memory['sums_x']/memory['sums_n'];
            break;
        case 'stats_my':
            x = memory['sums_y']/memory['sums_n'];
            break;
        case 'stats_wy':
            x = memory['sums_xy']/memory['sums_y'];
            break;
        case 'stats_sx':
            x  = Math.sqrt(memory['sums_x2']/memory['sums_n'] - Math.pow(memory['sums_x']/memory['sums_n'],2));
            x *= Math.sqrt(memory['sums_n']/(memory['sums_n']-1));
            break;
        case 'stats_sy':
            x  = Math.sqrt(memory['sums_y2']/memory['sums_n'] - Math.pow(memory['sums_y']/memory['sums_n'],2));
            x *= Math.sqrt(memory['sums_n']/(memory['sums_n']-1));
            break;
        case 'stats_r':
            // Pearson correlation coefficient
            x = stats_r();
            break;
        case 'stats_m':
            // least-squares slope
            x = stats_m();
            break;
        case 'stats_b':
            // least-squares y-intercept
            x = stats_b();
            break;
        default:
            console.log('unhandled op10');
            break;
    }

    rets.push(x);
    return true;
}

// 0 return value
// 1 argument
function op01(rets, args, operator)
{
    var x = args[0];

    switch(operator)
    {
        case 'srand':
            Math.seedrandom(x);
            break;
        case 'qmax':
            if (x <= 2 || x >= 1e6)
                return false;
            set_memory('qmax', Math.floor(x));
            break;
        default:
            console.log('unhandled op01');
            break;
    }

    return true;
}

// 1 return value
// 1 argument
function op11(rets, args, operator)
{
    var x = args[0];

    switch(operator)
    {
        case 'sqrt':
            if (x < 0)
                return false;
            x = Math.sqrt(x);
            break;
        case 'square':
            x = x*x;
            break;
        case 'exp':
            x = Math.exp(x);
            break;
        case 'pow10':
            x = Math.pow(10,x);
            break;
        case 'ln':
            if (x <= 0)
                return false;
            x = Math.log(x);
            break;
        case 'log10':
            if (x <= 0)
                return false;
            x = Math.log10(x);
            break;
        case 'inv':
            if (x === 0)
                return false;
            x = 1/x;
            break;
        case '!':
            if (x < 0)
                return false;
            x = Math.factorial(x);
            break;
        case 'rnd':
            x = Math.round(x);
            break;

        case 'sin':
            x = in_radians() ? Math.sin(x) : Math.sin(Math.deg2rad(x));
            break;
        case 'asin':
            x = in_radians() ? Math.asin(x) : Math.rad2deg(Math.asin(x));
            break;
        case 'cos':
            x = in_radians() ? Math.cos(x) : Math.cos(Math.deg2rad(x));
            break;
        case 'acos':
            x = in_radians() ? Math.acos(x) : Math.rad2deg(Math.acos(x));
            break;
        case 'tan':
            x = in_radians() ? Math.tan(x) : Math.tan(Math.deg2rad(x));
            break;
        case 'atan':
            x = in_radians() ? Math.atan(x) : Math.rad2deg(Math.atan(x));
            break;

        case 'hyp_sin':
            x = Math.sinh(x);
            break;
        case 'hyp_asin':
            x = Math.asinh(x);
            break;
        case 'hyp_cos':
            x = Math.cosh(x);
            break;
        case 'hyp_acos':
            if (x < 1)
                return false;
            x = Math.acosh(x);
            break;
        case 'hyp_tan':
            x = Math.tanh(x);
            break;
        case 'hyp_atan':
            if (Math.abs(x) >= 1)
                return false;
            x = Math.atanh(x);
            break;

        case 'frac':
            x = x - Math.toInt(x);
            break;
        case 'ip':
            x = Math.toInt(x);
            break;
        case 'to_deg':
            x = Math.rad2deg(x);
            break;
        case 'to_rad':
            x = Math.deg2rad(x);
            break;
        case 'to_m':
            x = 0.9144 * x;
            break;
        case 'to_yds':
            x = x / 0.9144;
            break;
        case 'to_ml':
            x = 29.5735295625 * x;
            break;
        case 'to_oz':
            x = x / 29.5735295625;
            break;
        case 'to_km':
            x = 1.609344 * x;
            break;
        case 'to_mi':
            x = x / 1.609344;
            break;
        case 'to_kg':
            x = 0.45359237 * x;
            break;
        case 'to_lb':
            x = x / 0.45359237;
            break;
        case 'to_l':
            x = 3.785411784 * x;
            break;
        case 'to_gal':
            x = x / 3.785411784;
            break;
        case 'to_cm':
            x = 2.54 * x;
            break;
        case 'to_in':
            x = x / 2.54;
            break;
        case 'to_degc':
            x = (x - 32) * 5/9;
            break;
        case 'to_degf':
            x = x * 9/5 + 32;
            break;

        case 'not':
            x = ~x;
            break;
        case '2cmp':
            x = ~x+1;
            break;

        case 'stats_xhat':
            x = (x - stats_b())/stats_m();
            break;
        case 'stats_yhat':
            x = stats_m()*x + stats_b();
            break;

        case 'to_hr':
            if (x < 0)
                return false;
            var min, sec;
            min = Math.floor(100*(x - Math.floor(x)));
            sec = 100*(100*x - Math.floor(100*x));
            x = Math.floor(x) + min/60 + sec/3600;
            break;
        case 'to_hms':
            if (x < 0)
                return false;
            min = 60*(x - Math.floor(x));
            sec = 60*(min - Math.floor(min));
            x = Math.floor(x) + Math.floor(min)/100 + sec/10000;
            break;

        default:
            console.log('unhandled op11');
            break;
    }

    rets.push(x);
    return true;
}

// 0 return values
// 2 arguments
function op02(rets, args, operator)
{
    var y = args[0];
    var x = args[1];

    switch(operator)
    {
        case 'sigma+':
            memory['sums_n']++;
            memory['sums_x']  += x;
            memory['sums_y']  += y;
            memory['sums_xy'] += x*y;
            memory['sums_x2'] += x*x;
            memory['sums_y2'] += y*y;
            memory_save_sums();
            break;
        case 'sigma-':
            memory['sums_n']--;
            memory['sums_x']  -= x;
            memory['sums_y']  -= y;
            memory['sums_xy'] -= x*y;
            memory['sums_x2'] -= x*x;
            memory['sums_y2'] -= y*y;
            memory_save_sums();
            break;
        default:
            console.log('unhandled op02');
            break;
    }

    return true;
}

// 1 return value
// 2 arguments
function op12(rets, args, operator)
{
    var y = args[0];
    var x = args[1];

    switch(operator)
    {
        case 'pow':
            x = Math.pow(y,x);
            break;
        case 'root':
            x = Math.pow(y,1/x);
            break;
        case '%':
            x = y * x/100;
            break;
        case '%chg':
            x = (x-y)/y*100;
            break;
        case 'Pnk':
            if (y < x || Math.floor(x) !== x || Math.floor(y) !== y)
                return false;
            x = Math.permutations(y,x);
            break;
        case 'Cnk':
            if (y < x || Math.floor(x) !== x || Math.floor(y) !== y)
                return false;
            x = Math.combinations(y,x);
            break;
        case 'div':
            if (x === 0)
                return false;
            x = y/x;
            break;
        case 'mul':
            x = y*x;
            break;
        case 'sub':
            x = y-x;
            break;
        case 'add':
            x = y+x;
            break;
        case 'and':
            x = x&y;
            break;
        case 'nand':
            x = ~(x&y)
            break;
        case 'or':
            x = x|y;
            break;
        case 'nor':
            x = ~(x|y);
            break;
        case 'xor':
            x = x^y;
            break;
        case 'xnor':
            x = ~(x^y);
            break;
        case 'mod':
            if (x === 0)
                return false;
            x = y%x;
            break;
        case 'shl':
            x = y<<x;
            break;
        case 'shr':
            x = y>>x;
            break;
        default:
            console.log('unhandled op12');
            break;
    }

    rets.push(x);
    return true;
}

// 2 return values
// 2 arguments
function op22(rets, args, operator)
{
    var y = args[0];
    var x = args[1];

    var xo, yo, zo;

    switch(operator)
    {
        case 'xchg':
            xo = y;
            yo = x;
            break;

        case 'complex_sign':
            xo = -x;
            yo = -y;
            break;
        case 'complex_sqrt':
            zo = Math.complex_pow([x,y], [0.5,0]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_square':
            zo = Math.complex_mul([x,y], [x,y]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_exp':
            zo = Math.complex_exp([x,y]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_ln':
            zo = Math.complex_ln([x,y]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_inv':
            zo = Math.complex_div([1,0],[x,y]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_sin':
            zo = Math.complex_sin([x,y]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_cos':
            zo = Math.complex_cos([x,y]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_tan':
            zo = Math.complex_tan([x,y]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_asin':
            zo = Math.complex_asin([x,y]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_acos':
            zo = Math.complex_acos([x,y]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_atan':
            zo = Math.complex_atan([x,y]);
            xo = zo[0]; yo = zo[1];
            break;

        case 'to_polar':
            yo = in_radians() ? Math.atan2(y,x) : Math.rad2deg(Math.atan2(y,x));
            xo = Math.sqrt(x*x + y*y);
            break;
        case 'to_cartesian':
            xo = in_radians() ? x*Math.cos(y) : x*Math.cos(Math.deg2rad(y));
            yo = in_radians() ? x*Math.sin(y) : x*Math.sin(Math.deg2rad(y));
            break;

        default:
            console.log('unhandled op22');
            break;
    }

    rets.push(yo);
    rets.push(xo);
    return true;
}

// 2 return values
// 4 arguments
function op24(rets, args, operator)
{
    var y2 = args[0];        // (x2,y2) was pushed first
    var x2 = args[1];
    var y1 = args[2];
    var x1 = args[3];
    var xo, yo, zo;

    switch(operator)
    {
        case 'complex_pow':
            zo = Math.complex_pow([x2,y2], [x1,y1]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_root':
            zo = Math.complex_pow([x2,y2], Math.complex_inv([x1,y1]));
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_div':
            zo = Math.complex_div([x2,y2], [x1,y1]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_mul':
            zo = Math.complex_mul([x2,y2], [x1,y1]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_sub':
            zo = Math.complex_sub([x2,y2], [x1,y1]);
            xo = zo[0]; yo = zo[1];
            break;
        case 'complex_add':
            zo = Math.complex_add([x2,y2], [x1,y1]);
            xo = zo[0]; yo = zo[1];
            break;

        default:
            console.log('unhandled op24');
            break;
    }

    rets.push(yo);
    rets.push(xo);
    return true;
}

function key_pressed_alt_l(pressed)
{
    if (pressed)
    {
        if (vibra_press.start)
            vibra_press.start();

        alt_l = !button_alt_l.checked;
        alt.text = alt_l ? "↰" : "";
        alt.color = "#ffbb00"
        if (alt_l)
            alt_r = button_alt_r.checked = false;
    }
}

function key_pressed_alt_r(pressed)
{
    if (pressed)
    {
        if (vibra_press.start)
            vibra_press.start();

        alt_r = !button_alt_r.checked;
        alt.text = alt_r ? "↱" : "";
        alt.color = "#00bbff"
        if (alt_r)
            alt_l = button_alt_l.checked = false;
    }
}

function key_pressed_c(pressed)
{
    if (pressed)
    {
        if (vibra_press.start)
            vibra_press.start();
    }
    else
    {
        if (vibra_release.start)
            vibra_release.start();

        if (alt_l || alt_r)
            Qt.quit();
        else
        {
            clear();
            update_display();
        }
    }
}

function try_state_change(new_state, nargs)
{
    if (!argument_check(nargs))
        return;
    state = new_state;
}

function key_pressed_dn(row, col)
{
    // SHOW (alt_r + ENTER)
    if (row === 2 && col === 0 && alt_r)
    {
        var prev_display_width = option['display_width'];
        option['display_width'] = 20;
        update_display();
        option['display_width'] = prev_display_width;
    }

    // FRAC (alt_l + .)
    if (row === 6 && col === 2 && alt_l)
    {
        option['display_rational'] = true;
        update_display();
        option['display_rational'] = false;
    }
}

function key_pressed_up(row, col)
{
    var h = key_handler[row][col];
    var f = null;

    if (CalculatorState.Reset <= state && state <= CalculatorState.DispSetup)
    {
        if (is_real())
        {
            if (alt_l)
                f = h.r_alt_l;
            else if (alt_r)
                f = h.r_alt_r;
            else if (!alt_l && !alt_r)
                f = h.r_main;
        }
        else
        {
            if (alt_l)
                f = h.i_alt_l;
            else if (alt_r)
                f = h.i_alt_r;
            else if (!alt_l && !alt_r)
                f = h.i_main;
        }

        if (!f)
            print_error();
        else
            f();
    }
    else if (CalculatorState.AlphaEntrySTO <= state && state <= CalculatorState.AlphaSwapY)
    {
        var a, x, y;

        if (!h.alpha)
            print_error();
        else
        {
            if (state === CalculatorState.AlphaEntrySTO)
                set_memory(h.alpha, stack.top());
            else if (state === CalculatorState.AlphaEntryRCL)
            {
                x = (memory[h.alpha] === null) ? 0 : parseFloat(memory[h.alpha]);
                history_update([], [x]);
                stack.push(x);
            }
            else if (state === CalculatorState.AlphaSwapX)
            {
                x = stack.pop();
                a = (memory[h.alpha] === null) ? 0 : parseFloat(memory[h.alpha]);
                set_memory(h.alpha, x);

                history_update([x], [a]);
                stack.push(a);
            }
            else if (state === CalculatorState.AlphaSwapY)
            {
                x = stack.pop();
                y = stack.pop();
                a = (memory[h.alpha] === null) ? 0 : parseFloat(memory[h.alpha]);
                set_memory(h.alpha, y);

                history_update([y,x], [a,x]);
                stack.push(a);
                stack.push(x);
            }

            state = CalculatorState.Reset;
        }
    }
    else if (state === CalculatorState.ComplexFunction)
    {
        state = CalculatorState.Reset;

        if (alt_l)
            f = h.c_alt_l;
        else if (alt_r)
            f = h.c_alt_r;
        else if (!alt_l && !alt_r)
            f = h.c_main;

        if (!f)
            print_error();
        else
            f();
    }
    else if (state === CalculatorState.HyperbolicFunction)
    {
        state = CalculatorState.Reset;

        if (alt_l)
            f = h.h_alt_l;
        else if (alt_r)
            f = h.h_alt_r;
        else if (!alt_l && !alt_r)
            f = h.h_main;

        if (!f)
            print_error();
        else
            f();
    }

    alt_l = button_alt_l.checked = 0;
    alt_r = button_alt_r.checked = 0;
    alt.text = "";

    update_display();
}

function key_pressed(pressed, row, col)
{
    if (pressed)
    {
        if (vibra_press.start)
            vibra_press.start();
        key_pressed_dn(row, col);
    }
    else
    {
        if (vibra_release.start)
            vibra_release.start();
        key_pressed_up(row, col);
    }
}

function initialize_callbacks()
{
    var rows = 7, cols = 6;

    key_handler = new Array(rows);
    for (var i = 0; i < rows; i++)
        key_handler[i] = new Array(cols);

    // sqrt
    key_handler[0][0] = new Object
    key_handler[0][0].r_alt_l = function(){ op(1,1,'square')  };
    key_handler[0][0].r_alt_r = function(){ op(1,2,'Cnk')     };
    key_handler[0][0].r_main  = function(){ op(1,1,'sqrt')    };
    key_handler[0][0].i_main  = function(){ append_digit('A') };
    key_handler[0][0].c_alt_l = function(){ op(2,2,'complex_square') };
    key_handler[0][0].c_main  = function(){ op(2,2,'complex_sqrt') };
    key_handler[0][0].alpha   = 'A';

    // exp
    key_handler[0][1] = new Object
    key_handler[0][1].r_alt_l = function(){ op(1,1,'pow10')   };
    key_handler[0][1].r_alt_r = function(){ op(1,2,'Pnk')     };
    key_handler[0][1].r_main  = function(){ op(1,1,'exp')     };
    key_handler[0][1].i_main  = function(){ append_digit('B') };
    key_handler[0][1].c_main  = function(){ op(2,2,'complex_exp') };
    key_handler[0][1].alpha   = 'B';

    // ln
    key_handler[0][2] = new Object
    key_handler[0][2].r_alt_l = function(){ op(1,1,'log10')   };
    key_handler[0][2].r_alt_r = function(){ lr_menu.open()    };
    key_handler[0][2].r_main  = function(){ op(1,1,'ln')      };
    key_handler[0][2].i_main  = function(){ append_digit('C') };
    key_handler[0][2].c_main  = function(){ op(2,2,'complex_ln') };
    key_handler[0][2].alpha   = 'C';

    // pow
    key_handler[0][3] = new Object
    key_handler[0][3].r_alt_l = function(){ op(1,2,'root')    };
    key_handler[0][3].r_alt_r = function(){ mean_menu.open()  };
    key_handler[0][3].r_main  = function(){ op(1,2,'pow')     };
    key_handler[0][3].i_main  = function(){ append_digit('D') };
    key_handler[0][3].c_alt_l = function(){ op(2,4,'complex_root') };
    key_handler[0][3].c_main  = function(){ op(2,4,'complex_pow') };
    key_handler[0][3].alpha   = 'D';

    // 1/x
    key_handler[0][4] = new Object
    key_handler[0][4].r_alt_l = function(){ op(1,1,'!')         };
    key_handler[0][4].r_alt_r = function(){ stddev_menu.open()  };
    key_handler[0][4].r_main  = function(){ op(1,1,'inv')       };
    key_handler[0][4].i_main  = function(){ append_digit('E')   };
    key_handler[0][4].c_main  = function(){ op(2,2,'complex_inv') };
    key_handler[0][4].alpha   = 'E';

    // Σ+
    key_handler[0][5] = new Object
    key_handler[0][5].r_alt_l = function(){ op(0,2,'sigma-')    };
    key_handler[0][5].r_alt_r = function(){ sums_menu.open()    };
    key_handler[0][5].r_main  = function(){ op(0,2,'sigma+')    };
    key_handler[0][5].i_main  = function(){ append_digit('F')   };
    key_handler[0][5].alpha   = 'F';

    // STO
    key_handler[1][0] = new Object
    key_handler[1][0].r_alt_l = function(){ try_state_change(CalculatorState.ComplexFunction,    0) };
    key_handler[1][0].r_alt_r = function(){ try_state_change(CalculatorState.HyperbolicFunction, 0) };
    key_handler[1][0].r_main  = function(){ try_state_change(CalculatorState.AlphaEntrySTO, 1) }
    key_handler[1][0].i_main  = function(){ try_state_change(CalculatorState.AlphaEntrySTO, 1) }
    key_handler[1][0].alpha   = 'G';

    // RCL
    key_handler[1][1] = new Object
    key_handler[1][1].r_alt_l = function(){ op(1,0,'rand')  };
    key_handler[1][1].r_alt_r = function(){ op(0,1,'srand') };
    key_handler[1][1].r_main  = function(){ try_state_change(CalculatorState.AlphaEntryRCL, 0) };
    key_handler[1][1].i_main  = function(){ try_state_change(CalculatorState.AlphaEntryRCL, 0) };
    key_handler[1][1].alpha   = 'H';

    // R↓
    key_handler[1][2] = new Object
    key_handler[1][2].r_alt_l = function(){ op(1,1,'rnd') };
    key_handler[1][2].r_alt_r = function(){ rotate_up()   };
    key_handler[1][2].r_main  = function(){ rotate_down() };
    key_handler[1][2].i_alt_r = function(){ rotate_up()   };
    key_handler[1][2].i_main  = function(){ rotate_down() };
    key_handler[1][2].alpha   = 'I';

    // SIN
    key_handler[1][3] = new Object
    key_handler[1][3].r_alt_l = function(){ op(1,1,'asin') };
    key_handler[1][3].r_alt_r = function(){ op(1,0,'pi')   };
    key_handler[1][3].r_main  = function(){ op(1,1,'sin')  };
    key_handler[1][3].c_alt_l = function(){ op(2,2,'complex_asin') };
    key_handler[1][3].c_main  = function(){ op(2,2,'complex_sin')  };
    key_handler[1][3].h_alt_l = function(){ op(1,1,'hyp_asin')  };
    key_handler[1][3].h_main  = function(){ op(1,1,'hyp_sin')  };
    key_handler[1][3].alpha   = 'J';

    // COS
    key_handler[1][4] = new Object
    key_handler[1][4].r_alt_l = function(){ op(1,1,'acos') };
    key_handler[1][4].r_alt_r = function(){ op(1,2,'%')    };
    key_handler[1][4].r_main  = function(){ op(1,1,'cos')  };
    key_handler[1][4].c_alt_l = function(){ op(2,2,'complex_acos') };
    key_handler[1][4].c_main  = function(){ op(2,2,'complex_cos')  };
    key_handler[1][4].h_alt_l = function(){ op(1,1,'hyp_acos') };
    key_handler[1][4].h_main  = function(){ op(1,1,'hyp_cos')  };
    key_handler[1][4].alpha   = 'K';

    // TAN
    key_handler[1][5] = new Object
    key_handler[1][5].r_alt_l = function(){ op(1,1,'atan') };
    key_handler[1][5].r_alt_r = function(){ op(1,2,'%chg') };
    key_handler[1][5].r_main  = function(){ op(1,1,'tan')  };
    key_handler[1][5].c_alt_l = function(){ op(2,2,'complex_atan') };
    key_handler[1][5].c_main  = function(){ op(2,2,'complex_tan')  };
    key_handler[1][5].h_alt_l = function(){ op(1,1,'hyp_atan') };
    key_handler[1][5].h_main  = function(){ op(1,1,'hyp_tan')  };
    key_handler[1][5].alpha   = 'L';

    // ENTER
    key_handler[2][0] = new Object
    key_handler[2][0].r_alt_l = function(){ lastx() };
    key_handler[2][0].r_alt_r = function(){         };
    key_handler[2][0].r_main  = function(){ enter() };
    key_handler[2][0].i_alt_l = function(){ lastx() };
    key_handler[2][0].i_alt_r = function(){         };
    key_handler[2][0].i_main  = function(){ enter() };
    key_handler[2][0].alpha   = 'M';

    // x <-> y
    key_handler[2][2] = new Object
    key_handler[2][2].r_alt_l = function(){ try_state_change(CalculatorState.AlphaSwapX, 1) };
    key_handler[2][2].r_alt_r = function(){ try_state_change(CalculatorState.AlphaSwapY, 2) };
    key_handler[2][2].r_main  = function(){ op(2,2,'xchg') };
    key_handler[2][2].i_alt_l = function(){ try_state_change(CalculatorState.AlphaSwapX, 1) };
    key_handler[2][2].i_alt_r = function(){ try_state_change(CalculatorState.AlphaSwapX, 2) };
    key_handler[2][2].i_main  = function(){ op(2,2,'xchg') };
    key_handler[2][2].alpha   = 'N';

    // +/-
    key_handler[2][3] = new Object
    key_handler[2][3].r_alt_l = function(){ open_menu('disp') };
    key_handler[2][3].r_alt_r = function(){ open_menu('mode') };
    key_handler[2][3].r_main  = function(){ change_sign()     };
    key_handler[2][3].i_alt_l = function(){ open_menu('disp') };
    key_handler[2][3].i_alt_r = function(){ open_menu('mode') };
    key_handler[2][3].i_main  = function(){ change_sign()     };
    key_handler[2][3].c_main  = function(){ op(2,2,'complex_sign') };
    key_handler[2][3].alpha   = 'O';

    // E
    key_handler[2][4] = new Object
    key_handler[2][4].r_alt_l = function(){ op(1,1,'frac') };
    key_handler[2][4].r_alt_r = function(){ history_redo() };
    key_handler[2][4].r_main  = function(){ append_exp10() };
    key_handler[2][4].i_alt_r = function(){ history_redo() };
    key_handler[2][4].alpha   = 'P';

    // backspace
    key_handler[2][5] = new Object
    key_handler[2][5].r_alt_l = function(){ clear_menu.open()  };
    key_handler[2][5].r_alt_r = function(){ history_undo()     };
    key_handler[2][5].r_main  = function(){ backspace()        };
    key_handler[2][5].i_alt_l = function(){ clear_menu.open()  };
    key_handler[2][5].i_alt_r = function(){ history_undo()     };
    key_handler[2][5].i_main  = function(){ backspace()        };

    // b⇄
    key_handler[3][0] = new Object
    key_handler[3][0].r_alt_l = function(){ open_menu('base') };
    key_handler[3][0].r_alt_r = function(){ op(1,1,'ip')      };
    key_handler[3][0].r_main  = function(){ lastb()           };
    key_handler[3][0].i_alt_l = function(){ open_menu('base') };
    key_handler[3][0].i_alt_r = function(){ op(1,1,'ip')      };
    key_handler[3][0].i_main  = function(){ lastb()           };

    // 7
    key_handler[3][1] = new Object
    key_handler[3][1].r_alt_l = function(){ op(2,2,'to_polar')     };
    key_handler[3][1].r_alt_r = function(){ op(2,2,'to_cartesian') };
    key_handler[3][1].r_main  = function(){ append_digit('7')      };
    key_handler[3][1].i_main  = function(){ append_digit('7')      };
    key_handler[3][1].alpha   = 'Q';

    // 8
    key_handler[3][2] = new Object
    key_handler[3][2].r_alt_l = function(){ op(1,1,'to_hr')   };
    key_handler[3][2].r_alt_r = function(){ op(1,1,'to_hms')  };
    key_handler[3][2].r_main  = function(){ append_digit('8') };
    key_handler[3][2].i_main  = function(){ append_digit('8') };
    key_handler[3][2].alpha   = 'R';

    // 9
    key_handler[3][3] = new Object
    key_handler[3][3].r_alt_l = function(){ op(1,1,'to_deg')  };
    key_handler[3][3].r_alt_r = function(){ op(1,1,'to_rad')  };
    key_handler[3][3].r_main  = function(){ append_digit('9') };
    key_handler[3][3].i_main  = function(){ append_digit('9') };
    key_handler[3][3].alpha   = 'S';

    // div
    key_handler[3][4] = new Object
    key_handler[3][4].r_alt_l = function(){ op(1,2,'and')  };
    key_handler[3][4].r_alt_r = function(){ op(1,2,'nand') };
    key_handler[3][4].r_main  = function(){ op(1,2,'div')  };
    key_handler[3][4].i_alt_l = function(){ op(1,2,'and')  };
    key_handler[3][4].i_alt_r = function(){ op(1,2,'nand') };
    key_handler[3][4].i_main  = function(){ op(1,2,'div')  };
    key_handler[3][4].c_main  = function(){ op(2,4,'complex_div')  };

    // 4
    key_handler[4][1] = new Object
    key_handler[4][1].r_alt_l = function(){ op(1,1,'to_m')    };
    key_handler[4][1].r_alt_r = function(){ op(1,1,'to_yds')  };
    key_handler[4][1].r_main  = function(){ append_digit('4') };
    key_handler[4][1].i_main  = function(){ append_digit('4') };
    key_handler[4][1].alpha   = 'T';

    // 5
    key_handler[4][2] = new Object
    key_handler[4][2].r_alt_l = function(){ op(1,1,'to_ml')   };
    key_handler[4][2].r_alt_r = function(){ op(1,1,'to_oz')   };
    key_handler[4][2].r_main  = function(){ append_digit('5') };
    key_handler[4][2].i_main  = function(){ append_digit('5') };
    key_handler[4][2].alpha   = 'U';

    // 6
    key_handler[4][3] = new Object
    key_handler[4][3].r_alt_l = function(){ op(1,1,'to_km')   };
    key_handler[4][3].r_alt_r = function(){ op(1,1,'to_mi')   };
    key_handler[4][3].r_main  = function(){ append_digit('6') };
    key_handler[4][3].i_main  = function(){ append_digit('6') };
    key_handler[4][3].alpha   = 'V';

    // mul
    key_handler[4][4] = new Object
    key_handler[4][4].r_alt_l = function(){ op(1,2,'or')  };
    key_handler[4][4].r_alt_r = function(){ op(1,2,'nor') };
    key_handler[4][4].r_main  = function(){ op(1,2,'mul') };
    key_handler[4][4].i_alt_l = function(){ op(1,2,'or')  };
    key_handler[4][4].i_alt_r = function(){ op(1,2,'nor') };
    key_handler[4][4].i_main  = function(){ op(1,2,'mul') };
    key_handler[4][4].c_main  = function(){ op(2,4,'complex_mul')  };

    // 1
    key_handler[5][1] = new Object
    key_handler[5][1].r_alt_l = function(){ op(1,1,'to_kg')   };
    key_handler[5][1].r_alt_r = function(){ op(1,1,'to_lb')   };
    key_handler[5][1].r_main  = function(){ append_digit('1') };
    key_handler[5][1].i_main  = function(){ append_digit('1') };
    key_handler[5][1].alpha   = 'W';

    // 2
    key_handler[5][2] = new Object
    key_handler[5][2].r_alt_l = function(){ op(1,1,'to_l')    };
    key_handler[5][2].r_alt_r = function(){ op(1,1,'to_gal')  };
    key_handler[5][2].r_main  = function(){ append_digit('2') };
    key_handler[5][2].i_main  = function(){ append_digit('2') };
    key_handler[5][2].alpha   = 'X';

    // 3
    key_handler[5][3] = new Object
    key_handler[5][3].r_alt_l = function(){ op(1,1,'to_cm')   };
    key_handler[5][3].r_alt_r = function(){ op(1,1,'to_in')   };
    key_handler[5][3].r_main  = function(){ append_digit('3') };
    key_handler[5][3].i_main  = function(){ append_digit('3') };
    key_handler[5][3].alpha   = 'Y';

    // sub
    key_handler[5][4] = new Object
    key_handler[5][4].r_alt_l = function(){ op(1,2,'xor')  };
    key_handler[5][4].r_alt_r = function(){ op(1,2,'xnor') };
    key_handler[5][4].r_main  = function(){ op(1,2,'sub')  };
    key_handler[5][4].i_alt_l = function(){ op(1,2,'xor')  };
    key_handler[5][4].i_alt_r = function(){ op(1,2,'xnor') };
    key_handler[5][4].i_main  = function(){ op(1,2,'sub')  };
    key_handler[5][4].c_main  = function(){ op(2,4,'complex_sub')  };

    // 0
    key_handler[6][1] = new Object
    key_handler[6][1].r_alt_l = function(){ op(1,1,'to_degc') };
    key_handler[6][1].r_alt_r = function(){ op(1,1,'to_degf') };
    key_handler[6][1].r_main  = function(){ append_digit('0') };
    key_handler[6][1].i_main  = function(){ append_digit('0') };
    key_handler[6][1].alpha   = 'Z';

    // .
    key_handler[6][2] = new Object
    key_handler[6][2].r_alt_l = function(){                  };
    key_handler[6][2].r_alt_r = function(){ op(0,1,'qmax')   };
    key_handler[6][2].r_main  = function(){ append_period()  };
    key_handler[6][2].i_main  = function(){ append_period()  };

    // mod
    key_handler[6][3] = new Object
    key_handler[6][3].r_alt_l = function(){ op(1,2,'shl') };
    key_handler[6][3].r_alt_r = function(){ op(1,2,'shr') };
    key_handler[6][3].r_main  = function(){ op(1,2,'mod') };
    key_handler[6][3].i_alt_l = function(){ op(1,2,'shl') };
    key_handler[6][3].i_alt_r = function(){ op(1,2,'shr') };
    key_handler[6][3].i_main  = function(){ op(1,2,'mod') };

    // add
    key_handler[6][4] = new Object
    key_handler[6][4].r_alt_l = function(){ op(1,1,'not')  };
    key_handler[6][4].r_alt_r = function(){ op(1,1,'2cmp') };
    key_handler[6][4].r_main  = function(){ op(1,2,'add')  };
    key_handler[6][4].i_alt_l = function(){ op(1,1,'not')  };
    key_handler[6][4].i_alt_r = function(){ op(1,1,'2cmp') };
    key_handler[6][4].i_main  = function(){ op(1,2,'add')  };
    key_handler[6][4].c_main  = function(){ op(2,4,'complex_add')  };
}
